/*
	SimpleView
	views some .tga .qrt .gif in individual windows

A cookie of fortune by /bin/fortune :
Disclaimer: "These opinions are my own, though for a small fee they be yours too."
		-- Dave Haynie

Something I don't understand : WHO'S RESPONSIBLE FOR THE LABELS ?!?
Some objects'destructors are said to free their label. But what if they were given
constant strings ??

*/

#include "SVApplication.h"
#include "SVWindow.h"
#include <app/AppDefs.h>
#include <interface/Bitmap.h>
#include <interface/Menu.h>
#include <interface/PopUpMenu.h>
#include <interface/MenuItem.h>
#include <interface/PopUpMenu.h>
#include <interface/Alert.h>
#include <storage/StorageDefs.h>
#include <storage/Database.h>
#include <storage/Store.h>
#include <storage/File.h>
#include <storage/Directory.h>
#include <storage/Record.h>
#include <storage/Volume.h>
#include <support/List.h>
#include <kernel/OS.h>
#include <inc/string.h>

#include <inc/stdio.h>

extern SVApplication *sv_app;

void error( char *err_string )
{
	BAlert *al = new BAlert( "SimpleView : error", err_string, "OK", NULL, NULL,
	 B_WIDTH_AS_USUAL, B_STOP_ALERT ); ASSERT( al!=0 );
	al->Go( );
}

/* SVApplication constructor :
We create an empty pictures (SVWindow's) list and our alternate app menu.
*/
SVApplication::SVApplication( ulong sig )
	: BApplication( sig )
{
// we set up an empty picture windows list
	pic_list = new BList( ); ASSERT( pic_list!=0 );
	received_something = FALSE;
	
	directory.database = 0;
	directory.record = 0;

// we set our own menu, with an open picture entry (shorcut alt-O) and a separator line.
	BPopUpMenu *main_menu = new BPopUpMenu( "SimpleView MainMenu", FALSE, FALSE ); ASSERT( main_menu!=0 );
	BMenuItem *item;
	BMessage *msg;
	BMenu *submenu;

	msg = new BMessage( B_ABOUT_REQUESTED ); ASSERT( msg!=0 );
	item = new BMenuItem( "About SimpleView", msg ); ASSERT( item!=0 );
	main_menu->AddItem( item );

	msg = new BMessage( OPEN_FILE ); ASSERT( msg!=0 );
	item = new BMenuItem( "Open a picture", msg, 'O' ); ASSERT( item!=0 );
	main_menu->AddItem( item );

// Empty on a first time
	submenu = new BMenu( "Pictures" ); ASSERT( submenu!=0 );
	main_menu->AddItem( submenu );

	bool ok = main_menu->AddSeparatorItem( ); ASSERT( ok==TRUE );

	msg = new BMessage( B_QUIT_REQUESTED ); ASSERT( msg!=0 );
	item = new BMenuItem( "Quit SimpleView", msg, 'Q' ); ASSERT( item!=0 );
	main_menu->AddItem( item );

	main_menu->SetTarget( this );	
	SetMainMenu( main_menu );
}

SVApplication::~SVApplication( )
{
	delete pic_list;
}

// no comment :-) 
void SVApplication::AboutRequested( )
{
	BAlert *about = new BAlert( "About SimpleView", "SimpleView 1.1, a \
simple image viewer for .targa/.tga and .qrt/.dump -- 96 Benoit Triquet", "OK" ); ASSERT( about!=0 );
	about->Go( );
}

/* ReadyToRun :
if we don't have at least one image to open in a window, we run the file panel
*/
void SVApplication::ReadyToRun( )
{
	if( NumPic( )==0 )
		if( !received_something )
			RunFilePanel( "SimpleView : Choose a picture", "Open picture" );
		else
			PostMessage( B_QUIT_REQUESTED );
}

/* UpdateSub( ) :
We rebuild the Pictures submenu which lists the opened windows
*/
void SVApplication::UpdateSub( )
{
	BMenuItem *item;
	BMessage *msg;
// we remove then delete the submenu. As I'm lazy we rebuild the entire submenu
	BMenu *submenu = MainMenu( )->SubmenuAt( 2l );
	MainMenu( )->RemoveItem( 2l );
	delete submenu;
	submenu = new BMenu( "Pictures" ); ASSERT( submenu!=0 );
	for( long i = 0; i<pic_list->CountItems( ); i++ ) {
		msg = new BMessage( ACTIVATE_PIC ); ASSERT( msg!=0 );
// note that we don't need to add the number of the window in the message :
// the BMenuItem adds itself an index entry and the indexes in the submenu match those
// in the pic_list
// *WARNING* perhaps it's better to add such, so that the right window gets activated !!!
// can't figure why the first way doesn't work. Anyway let's do it !
		msg->AddLong( "picnum", i );
		item = new BMenuItem( ( ( SVWindow * )pic_list->ItemAt( i ) )->Title( ), msg ); ASSERT( item!=0 );
		submenu->AddItem( item );
	}
// don't forget this one !!! or you'll search your messages for a loooong long time...
	submenu->SetTarget( this );
	MainMenu( )->AddItem( submenu, 2l );
}

/* MessageReceived :
We probably have received a menu selection msg (!)
*/
void SVApplication::MessageReceived( BMessage *msg )
{
	switch( msg->what ) {
	case DO_POPUP_MENU:
		BPoint point = msg->FindPoint( "where" );
		BMenuItem *sel;
		BMessage *copy;
		sel = MainMenu( )->Go( point );
		if( sel!=NULL && sel->Message( )!=NULL ) {
			copy = new BMessage( sel->Message( ) ); ASSERT( copy!=0 );
			PostMessage( copy );
		}
		break;
	case ACTIVATE_PIC:
		( ( SVWindow * )pic_list->ItemAt( msg->FindLong( "picnum" ) ) )->Activate(  );
		break;
	case OPEN_FILE:
		if( directory.database!=0 && directory.record!=0 ) {
			BMessage *conf_msg = new BMessage( B_REFS_RECEIVED ); ASSERT( conf_msg!=0 );
			conf_msg->AddRef( "directory", directory );
			RunFilePanel( "SimpleView : Open another picture", "Open picture", FALSE, conf_msg );
		} else
			RunFilePanel( "SimpleView : Open another picture", "Open picture" );
		break;
	default:
		BApplication::MessageReceived( msg );
	}
}

/* ArgvReceived :
*BAD* We have files on the command line : let's find their refs and post a msg to
*BAD* ourselves so that the command line is processed as the dropped files or the file panel.
THIS WON'T WORK !!!
Unfortunately this msg will arrive after the B_READY_TO_RUN, which will find an empty
pic_list whatever the command line add. So we must try to open the file right here, and
let ReadyToRun check AFTER.
*/
void SVApplication::ArgvReceived( int argc, char **argv )
{
	if( argc>0 ) {
		record_ref ref;
		BFile f;
		BDirectory d;
		for( int i = 1; i<argc; i++ ) { // < not <= !!! 3 days for that B:-P
			if( get_ref_for_path( argv[ i ], &ref )==B_NO_ERROR )
				if( ref.database>=0 && ref.record>=0 )
					if( does_ref_conform( ref, "File" ) ) {
						OpenPic( ref );
// more work than necessary, and i don't care
						f.SetRef( ref );
						f.GetParent( &d );
						directory = d.Record( )->Ref( );
					}
		}
	}
	received_something = TRUE;
}

/* RefsReceived :
We received a ref from the file panel, the app menu or the command line,
we check if it's really a file, and if so set the current directory ref on that file's
parent directory and pass the ref to OpenPic.
*/
void SVApplication::RefsReceived( BMessage *msg ) // probably load a new pic
{
// Another *BUG* : the file panel will close itself very well, IN NO CASE we close it ourselves.
// I don't even want to know what is the value of IsFilePanelRunning( )
//	if( IsFilePanelRunning( ) )
//		CloseFilePanel( );

// *BUG* ! directory here would receive the directory which we openned the panel on !
	if( msg->HasRef( "directory" ) )
		directory = msg->FindRef( "directory" );
	ulong type;
	long count;
	uint i;
	record_ref ref;
	BFile f;
	BDirectory d;
	msg->GetInfo( "refs", &type, &count );
	for( i = 0; i<count; i++ ) {
		ref = msg->FindRef( "refs", i );
		if( ref.database>=0 && ref.record>=0 )
			if( does_ref_conform( ref, "File" ) ) {
				OpenPic( ref );
// we set directory on the directory of the last file openned (be it successful or not)
				f.SetRef( ref );
				f.GetParent( &d );
				directory = d.Record( )->Ref( );
			}
	}
	if( received_something==FALSE && NumPic( )==0 )
		PostMessage( B_QUIT_REQUESTED );
	received_something = TRUE;
}

/* FilePanelClosed :
we set directory according to the msg "directory" field if the file panel hasn't been
canceled. If we had nothing on startup and got a cancel on the file panel, we must quit.
*/
void SVApplication::FilePanelClosed( BMessage *msg )
{
// the "canceled" field is a bool, whereas the bebook states it's a long
	if( ! msg->HasBool( "canceled" ) )
// in the facts this will NEVER happen !
		directory = msg->FindRef( "directory" );
	else
		if( received_something==FALSE )
			PostMessage( B_QUIT_REQUESTED );
}

/* OpenPic :
We open the file for the given ref, create a window which displays it and add an entry
in the application opened pictures list.
*/
void SVApplication::OpenPic( record_ref ref )
{
// check for file, open it and get its name
	BFile file;
	uint w, h;
	BBitmap *pic;
	file.SetRef( ref );
	if( file.OpenData( )!=B_NO_ERROR ) {
		error( "Unable to open the data portion of the file" );
		return ;
	}
//  *BAD*	name = file.Name( ); BStore::GetName hasn't been replaced by Name !!
	char name[ B_FILE_NAME_LENGTH ];
	file.GetName( name );

// identify on extension an read
	char *ext = strrchr( name, '.' );
	if( ext==0 ) {
		error( "Unable to identify a file name without extension" );
		return ;
	}
	char *p = ext;
	while( *( ++p )!=0 )
		if( *p>='A' && *p<='Z' )
			*p += 'a'-'A';
	if( strncmp( ext, ".tga", 4 )==0 || strncmp( ext, ".targa", 6 )==0 ) {
/* targa
header (18 bytes) :
+0	2 bytes		- idlen, picture identification length
+2	1 bytes		- ftype, 9 & 10 RLE compression, 1, 2 RGB24, 3 simple rasters
+3	2 bytes		- ?
+5	2 bytes		- cmlen, color map length in the file
+7	2 bytes		- cmsize, color map size
+8	4 bytes		- ?
+12	2 bytes		- width, little endian
+14	2 bytes		- height, little endian
+16	1 bytes		- psize, pixel size in bits, 24 for RGB24
+17	1 bytes		- bit 5 : orientation, 1 is top line before, 0 is bottom line before
If idlen!=0, identification (idlen bytes), a comment

FOR RGB24 ONLY:
cmlen=0  cmsize=0
each scanline (width*3 bytes) :
	width times bb gg rr
*/
		uchar b[ 18 ];
// try to get the width and height
		if( file.Read( &b, 18 )!=18 ) {
			error( "Targa file : bad header" );
			return ;
		}
		if( b[ 2 ]!=2 || ( ( b[ 6 ]<<8 )|b[ 5 ] )!=0 || b[ 16 ]!=24 || !( b[ 17 ]&0x20 ) ) {
			error( "Targa file : only non-mapped 24 bits uncompressed right side up format is supported" );
			return ;
		}
		uint idlen = ( b[ 1 ]<<8 )|b[ 0 ];
		if( file.Seek( idlen, 1 )!=18+idlen ) {
			error( "Targa file : bad identification field" );
			return ;
		}
		w = ( b[ 13 ]<<8 )|b[ 12 ];
		h = ( b[ 15 ]<<8 )|b[ 14 ];
// allocate a qrt format and a rgb_color format line, clear the rgb line alpha channel
		uchar *line = new uchar[ w*3 ]; ASSERT( line!=0 );
		int l = 0, c;
// allocate the BBitmap, clear it
		pic = new BBitmap( BRect( 0.0f, 0.0f, ( float )( w-1 ), ( float )( h-1 ) ), B_COLOR_8_BIT ); ASSERT( pic!=0 );
		memset( pic->Bits( ), 0, pic->BitsLength( ) );
// read the lines (as much as possible)
		while( file.Read( line, w*3 )==w*3 && l<h ) {
// convert a bgr line to rgb then copy it into the BBitmap with 24 to 8 bits conv.
// BE BUG : SetBits when converting from 24 to 8 bits reads 3 bytes per pixel, not 4
			uchar t;
			for( c = 0; c<w; c++ ) {
				t = line[ 3*c+2 ];
				line[ 3*c+2 ] = line[ 3*c ];
				line[ 3*c ] = t;
			}
			pic->SetBits( line, w*3, l*w, B_RGB_24_BIT );
			l++;
		}
		delete [ ] line;
	} else 
		if( strncmp( ext, ".qrt", 4 )==0 || strncmp( ext, ".dump", 5 )==0 ) {
/* .qrt / .dump
header (2 bytes) :
wwww hhhh		- Width, Height (16 bits, LSB first)
each scanline (2+Width*3 bytes) :
llll			- Line number (16 bits, LSB first)
rr rr rr ...	- Red data for line (8 bits per pixel,
				  left to right, 0-255 (255=bright, 0=dark))
gg gg gg ...	- Green data for line (8 bits per pixel,
				  left to right, 0-255 (255=bright, 0=dark))
bb bb bb ...	- Blue data for line (8 bits per pixel,
				  left to right, 0-255 (255=bright, 0=dark))
*/
			uchar b[ 4 ];
// try to get the width and height
			if( file.Read( &b, 4 )!=4 ) {
				error( "QRT file : bad header" );
				return ;
			}
			w = ( b[ 1 ]<<8 )|b[ 0 ];
			h = ( b[ 3 ]<<8 )|b[ 2 ];
// allocate a qrt format and a rgb_color format line, clear the rgb line alpha channel
			uchar *sline = new uchar[ w*3 ]; ASSERT( sline!=0 );
			uchar *dline = new uchar[ w*3 ]; ASSERT( dline!=0 );
			int l = 0, read_l, c;
// allocate the BBitmap, clear it
			pic = new BBitmap( BRect( 0.0f, 0.0f, ( float )( w-1 ), ( float )( h-1 ) ), B_COLOR_8_BIT ); ASSERT( pic!=0 );
			memset( pic->Bits( ), 0, pic->BitsLength( ) );
// read the lines (as much as possible)
			while( file.Read( &b, 2 )==2 && file.Read( sline, w*3 )==w*3 && l<h ) {
				read_l = ( b[ 1 ]<<8 )|b[ 0 ]; // get the line number from the file, test it
				if( l!=read_l ) {
					char err[ 80 ];
					sprintf( err, "QRT file : bad line number, read %d for line %d", read_l, l );
					error( err );
					delete pic;
					delete [ ] sline;
					delete [ ] dline;
					return ;
				}
// convert a qrt line to rgb format then copy it into the BBitmap with 24 to 8 bits conv.
// BE BUG : SetBits when converting from 24 to 8 bits reads 3 bytes per pixel, not 4
				for( c = 0; c<w; c++ ) {
					dline[ 3*c ] = sline[ c ];
					dline[ 3*c+1 ] = sline[ c+w ];
					dline[ 3*c+2 ] = sline[ c+2*w ];
				}
				pic->SetBits( dline, w*3, l*w, B_RGB_24_BIT );
				l++;
			}
			delete [ ] sline;
			delete [ ] dline;
		} else {
			error( "Unknown format, bad extension" );
			return ;
		}

	char *wname = new char[ B_FILE_NAME_LENGTH+3+2*5 ]; ASSERT( wname !=0 );
	sprintf( wname, "%s, %dx%d", name, w, h );
	SVWindow *win = new SVWindow( wname, pic ); ASSERT( win!= 0); // we don't delete wname !
	AddPic( win );
}

//  Benot Triquet 1996